# aclnnEluBackward

## 支持的产品型号
- Atlas 推理系列产品（Ascend 310P处理器）。
- Atlas 训练产品。
- Atlas A2训练系列产品。

## 接口原型

每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnEluBackwardGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnEluBackward”接口执行计算。

* `aclnnStatus aclnnEluBackwardGetWorkspaceSize(const aclTensor *gradOutput, const aclScalar *alpha, const aclScalar *scale, const aclScalar *inputScale, bool isResult, const aclTensor *gradOutputOrResult, aclTensor *gradInput, uint64_t *workspaceSize, aclOpExecutor **executor)`
* `aclnnStatus aclnnEluBackward(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, aclrtStream stream)`

## 功能描述

- 算子功能：完成ELU激活函数的反向计算，输出ELU激活函数正向输入的梯度。

- 计算公式：x是gradOutputOrResult中的某个元素。

1. 当isResult是True时：

$$
gradInput = gradOutput *
\begin{cases}
scale, \quad x > 0\\
\alpha \ast scale \ast inputScale \ast (x+1), \quad x \leq 0
\end{cases}
$$

2. 当isResult是False时：

$$
gradInput = gradOutput *
\begin{cases}
scale, \quad x > 0\\
\alpha \ast scale \ast inputScale \ast exp(x \ast inputScale), \quad x \leq 0
\end{cases}
$$

## aclnnEluBackwardGetWorkspaceSize

- **参数说明：**

  - gradOutput(aclTensor*,计算输入)：表示ELU激活函数正向输出的梯度，可自定义，类似于权重，Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BFLOAT16（仅Atlas A2训练系列产品支持），支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，数据维度不支持8维以上。

  - alpha(aclTensor*,计算输入)：表示ELU激活函数的激活系数，Host侧的aclScalar，数据类型需要是可转换为FLOAT的数据类型。

  - scale(aclTensor*,计算输入)：表示ELU激活函数的缩放系数，Host侧的aclScalar，数据类型需要是可转换为FLOAT的数据类型。

  - inputScale(aclTensor*,计算输入)：表示ELU激活函数的输入的缩放系数，Host侧的aclScalar，数据类型需要是可转换为FLOAT的数据类型。

  - isResult(BOOL,计算输入)：表示传给ELU反向计算的输入是否是ELU正向的输出，数据类型支持BOOL。

  - gradOutputOrResult(aclTensor*,计算输入)：当isResult为True时，表示ELU激活函数正向的输出，当isResult为False时，表示ELU激活函数正向的输入，Device侧的aclTensor，数据类型支持FLOAT、FLOAT16，数据类型需要与gradOutput一致。shape需要和gradOutput的shape一致，支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，数据维度不支持8维以上。

  - gradInput(aclTensor*,计算输出)：表示ELU激活函数正向输入的梯度，即对输入进行求导后的结果，Device侧的aclTensor，数据类型需要是gradOutput可转换的数据类型，shape需要和gradOutput的shape一致，支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，数据维度不支持8维以上。

  - workspaceSize(uint64_t*,出参)：返回用户需要在Device侧申请的workspace大小。

  - executor(aclOpExecutor**,出参)：返回op执行器，包含了算子计算流程。


- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。
  ```
  161001(ACLNN_ERR_PARAM_NULLPTR)：1. 参数gradOutput、alpha、scale、inputScale、gradOutputOrResult、gradInput是空指针。
  161002(ACLNN_ERR_PARAM_INVALID)：1. 参数gradOutput、gradOutputOrResult的数据类型不在支持的范围之内。
                                   2. 参数gradOutput、gradOutputOrResult的数据类型不一致。
                                   3. 参数alpha、scale、inputScale的数据类型不可转换为FLOAT。
                                   4. 参数gradInput的数据类型不是gradOutput可转换的。
                                   5. 参数gradOutput、gradOutputOrResult、gradInput的shape不一致。
                                   6. 参数gradOutput、gradOutputOrResult、gradInput的维度大于8。
                                   7. 参数isResult为True时，alpha小于0。
  ```

## aclnnEluBackward

- **参数说明：**

  * workspace(void*,入参)：在Device侧申请的workspace内存起址。

  * workspaceSize(uint64_t,入参)：在Device侧申请的workspace大小，由第一段接口aclnnEluBackwardGetWorkspaceSize获取。

  * executor(aclOpExecutor*,入参)：op执行器，包含了算子计算流程。

  * stream(aclrtStream,入参)：指定执行任务的 AscendCL Stream流。


- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无。

## 调用示例

示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。

```c++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_elu_backward.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** selfOrResult) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续selfOrResult的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *selfOrResult = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> gradOutputShape = {2, 2};
  std::vector<int64_t> selfOrResultShape = {2, 2};
  std::vector<int64_t> gradInputShape = {2, 2};
  void* gradOutputDeviceAddr = nullptr;
  void* selfOrResultDeviceAddr = nullptr;
  void* gradInputDeviceAddr = nullptr;
  aclTensor* gradOutput = nullptr;
  aclScalar* alpha = nullptr;
  aclScalar* scale = nullptr;
  aclScalar* inputScale = nullptr;
  aclTensor* selfOrResult = nullptr;
  aclTensor* gradInput = nullptr;
  std::vector<float> gradOutputHostData = {-2, -1, 0, 1};
  std::vector<float> selfOrResultHostData = {-2, -1, 0, 1};
  std::vector<float> gradInputHostData = {0, 0, 0, 0};
  float alphaValue = 1.0f;
  float scaleValue = 1.0f;
  float inputScaleValue = 1.0f;
  bool isResult = true;
  // 创建gradOutput aclTensor
  ret = CreateAclTensor(gradOutputHostData, gradOutputShape, &gradOutputDeviceAddr, aclDataType::ACL_FLOAT,
                        &gradOutput);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建alpha aclScalar
  alpha = aclCreateScalar(&alphaValue, aclDataType::ACL_FLOAT);
  CHECK_RET(alpha != nullptr, return ret);
  // 创建scale aclScalar
  scale = aclCreateScalar(&scaleValue, aclDataType::ACL_FLOAT);
  CHECK_RET(scale != nullptr, return ret);
  // 创建inputScale aclScalar
  inputScale = aclCreateScalar(&inputScaleValue, aclDataType::ACL_FLOAT);
  CHECK_RET(inputScale != nullptr, return ret);
  // 创建selfOrResult aclTensor
  ret = CreateAclTensor(selfOrResultHostData, selfOrResultShape, &selfOrResultDeviceAddr, aclDataType::ACL_FLOAT,
                        &selfOrResult);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建gradInput aclTensor
  ret = CreateAclTensor(gradInputHostData, gradInputShape, &gradInputDeviceAddr, aclDataType::ACL_FLOAT, &gradInput);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnEluBackward第一段接口
  ret = aclnnEluBackwardGetWorkspaceSize(gradOutput, alpha, scale, inputScale, isResult, selfOrResult, gradInput,
                                         &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnEluBackwardGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnEluBackward第二段接口
  ret = aclnnEluBackward(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnEluBackward failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(gradInputShape);
  std::vector<float> resultData(size, 0);
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), gradInputDeviceAddr,
                    size * sizeof(resultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(gradOutput);
  aclDestroyScalar(alpha);
  aclDestroyScalar(scale);
  aclDestroyScalar(inputScale);
  aclDestroyTensor(selfOrResult);
  aclDestroyTensor(gradInput);

  // 7. 释放device资源，需要根据具体API的接口定义修改
  aclrtFree(gradOutputDeviceAddr);
  aclrtFree(selfOrResultDeviceAddr);
  aclrtFree(gradInputDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  return 0;
}
```

