# aclnnResize
## 支持的产品型号
- Atlas 推理系列产品（Ascend 310P处理器）
- Atlas 训练产品
- Atlas A2训练系列产品

## 接口原型
每个算子分为两段式接口，必须先调用“aclnnResizeGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnResize”接口执行计算。

- `aclnnStatus aclnnResizeGetWorkspaceSize(const aclTensor *self, const aclFloatArray *scales, const char *mode, aclTensor *out, uint64_t *workspaceSize, aclOpExecutor **executor);`
- `aclnnStatus aclnnResize(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, aclrtStream stream);`

## 功能描述
算子功能：根据scales调整输入张量的大小
计算公式：
```
output_dimension = floor(input_dimension * scale)
```

## 接口参数
### aclnnResizeGetWorkspaceSize
- **参数说明：**
- self（const aclTensor*，计算输入）：含义说明，输入的张量，Device侧的aclTensor，数据类型支持FLOAT16, FLOAT，支持非连续的Tensor，数据格式支持NCHW。
- scales（const aclFloatArray *，计算输入）：含义说明，指定self张量调整的倍数，数据类型为float，长度和self维度相同。
- mode（const char *，属性）：含义说明，插值模式，只支持nearest或bilinear。
- out（aclTensor *，输出）：含义说明，计算输出，shape符合output_dimension = floor(input_dimension * scale)
- workspaceSize(uint64_t*)：返回用户需要在npu device侧申请的workspace大小。
- executor(aclOpExecutor**):返回op执行器，包含了算子计算流程。
- 返回值：aclnnStatus：返回状态码，具体参见aclnn返回码。
```
161001：传入的self、scales、mode、out中有空指针
161002：1. 数据类型不在支持的范围之内;
        2. shape不满足要求：
561101：内部异常：API内部创建aclOpExecutor失败（可能因为操作系统异常）。
561103：内部异常：aclnn API内部发生异常，出现了nullptr的异常。
```
### aclnnResize
- **参数说明：**
- workspace(void *, 入参): 在Device侧申请的workspace内存起址。
- workspaceSize(uint64_t, 入参): 在Device侧申请的workspace大小，由第一段接口aclnnScatterNdGetWorkspaceSize获取。
- executor(aclOpExecutor *, 入参): op执行器，包含了算子计算流程。
- stream(const aclrtStream, 入参): 指定执行任务的 AscendCL Stream流。
- 返回值：
 - aclnnStatus：返回状态码，具体参见aclnn返回码。
## 调用示例
```
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_resize.h"
#include <unistd.h>

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，acl初始化
  auto ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);

  ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);

  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_NCHW,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化, 参考acl对外接口列表
  // 根据自己的实际device填写deviceId

  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  // check根据自己的需要处理
  CHECK_RET(ret == 0, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> selfShape = {1,1,2,2};
  std::vector<int64_t> outShape = {1,1,4,4};

  void* selfDeviceAddr = nullptr;
  void* outDeviceAddr = nullptr;

  aclTensor* self = nullptr;
  aclTensor* out = nullptr;

  std::vector<float> selfHostData = {1.0,2.0,3.0,4.0};
  std::vector<float> scalesData = {1.0, 1.0, 2.0, 2.0};
  std::vector<float> outHostData(16);

  // 创建self aclTensor
  ret = CreateAclTensor(selfHostData, selfShape, &selfDeviceAddr, aclDataType::ACL_FLOAT, &self);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 创建out aclTensor
  ret = CreateAclTensor(outHostData, outShape, &outDeviceAddr, aclDataType::ACL_FLOAT, &out);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  aclFloatArray *scales = nullptr;
  scales = aclCreateFloatArray(scalesData.data(),scalesData.size());
  CHECK_RET(scales != nullptr, return 0);

  // 3. 调用CANN算子库API，需要修改为具体的HostApi
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  const char* mode = "nearest";
  // 调用aclnnResize第一段接口
  ret = aclnnResizeGetWorkspaceSize(self, scales, mode, out, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnResizeGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret;);
    printf("workspaceSize: %ld\n",workspaceSize);
  }
  // 调用aclnnResize第二段接口
  ret = aclnnResize(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnResizefailed. ERROR: %d\n", ret); return ret);
  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);
  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(outShape);
  std::vector<float> resultData(size, 0);
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), outDeviceAddr, size * sizeof(float),
                    ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(self);
  aclDestroyTensor(out);
  aclDestroyFloatArray(scales);

  // 7. 释放device资源，需要根据具体API的接口定义修改
  aclrtFree(selfDeviceAddr);
  aclrtFree(outDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  sleep(10);
  return 0;
}
```